`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date: 09/23/2023 03:41:41 PM
// Design Name: 
// Module Name: top
// Project Name: 
// Target Devices: 
// Tool Versions: 
// Description: 
// 
// Dependencies: 
// 
// Revision:
// Revision 0.01 - File Created
// Additional Comments:
// 
//////////////////////////////////////////////////////////////////////////////////


module top(
    input clock,    // system clock
    input reset,    // BTNR
    input version,  // BTNU
    input adc_n, adc_p, // VCAUX 6, P and N
    input [15:0] sw,    // slide switches
    output [15:0] led,  // 16  onboard LEDs above the switches
    output [6:0] segment,   // 4 digit LED display
    output dp,              // "." next to each LED digit
    output [3:0] digit,
    output reg [7:0] JB,
    output reg [7:0] JA,
    input rx,               // UART receive
    output tx               // UART transmit
    );
    parameter VERSION = 'h1001;
    //
    //  debounce the reset and version buttons
    //
    wire reset_level, reset_1s;
    debouncer DEBOUNCE_RESET(
        .clock(clock),
        .button(reset), 
        .ticks(17'd25000),      // 25k x 40ns = 1ms
        .level(reset_level), 
        .pulse(reset_1s)
    );
    wire version_level, version_1s;
    debouncer DEBOUNCE_VERSION(
        .clock(clock), 
        .button(version), 
        .ticks(17'd25000),      // 25k x 40ns = 1ms
        .level(version_level),
        .pulse(version_1s)
    );
    //
    //  generate a 25MHz and 8MHz clock for the UART and FIFO and a 104MHz clock for the XADC
    //  so that we can digitize at 1MSps
    //
    wire locked12, locked3, clock25, clock8, clock104;
    clock_8_25 MYCLOCKS12 (
        .reset(reset_1s),
        .clk_in1(clock),
        .locked(locked12),
        .clk_out1(clock25),
        .clk_out2(clock8)
    );
    clock_104 MYCLOCK3 (
        .reset(reset_1s),
        .clk_in1(clock),
        .locked(locked3),
        .clk_out1(clock104)
    );
    //
    //  divide clock8 by 8 to get the 1MHz ADC clock
    //
    reg [2:0] count8;
    always @ (posedge clock8) count8 <= count8 + 1;
    wire clock1;
    BUFG clk1buf (.I(count8[2]), .O(clock1));
    //
    //  next drive the 4 7-segment displays
    //
    wire [15:0] display_this;
    display4 DISPLAY (
        .clk100(clock),
        .number(display_this),
        .digit(digit),
        .segments(segment),
        .period(dp)
        );
    //
    //  instantiate the UART receiver.  run with the 50MHz clock so that
    //  we can run the transmitter FSM at 100MHz and not get lost
    //
    wire dv;
    wire [7:0] rx_data;
    uart_rx RECEIVER (
        .i_Clocks_per_Bit('d25),
        .i_Clock(clock25),
        .i_Reset(reset),
        .i_Rx_Serial(rx),       // tied to FPGA rx output
        .o_Rx_DV(dv),
        .o_Rx_Byte(rx_data)
        );
    //
    //  instantiate the UART transmitter
    //
    wire tx_active, tx_done;
    wire do_transmit;
    wire [7:0] transmit_byte;
    uart_tx TRANSMITTER (
        .i_Clocks_per_Bit('d25),
        .i_Clock(clock25),
        .i_Reset(reset),
        .i_Tx_DV(do_transmit),
        .i_Tx_Byte(transmit_byte), 
        .o_Tx_Active(tx_active),
        .o_Tx_Serial(tx),       // tied to FPGA tx output
        .o_Tx_Done(tx_done)
        );
    //
    //  here is the XADC block
    //
    wire [6:0] daddr_in = 7'h16;
    wire adc_ready, isbusy, adc_data_ready, eos_out, alarm;
    wire [15:0] adc_data;
    wire [4:0] channel_out;
    XADC_IP XADC_INST (
        .daddr_in(7'h16),   // specifies vcaux6 pints to digitize
        .dclk_in(clock104),    // 50MHz clock
        .den_in(adc_ready), // tied to adc_ready, tells adc to convert, tieing causes continuous conversions
        .di_in(16'h0),      // to set the data to something, not used here
        .dwe_in(1'b0),      //  set to enable writing to di_in, which we don't want to do
        .vauxp6(adc_p),     //  positive input to digitize
        .vauxn6(adc_n),     //  negative input to digitize
        .busy_out(isbusy),  // tells you the adc is busy converting
        .channel_out(channel_out[4:0]), // for using more than 1 channel, tells you which one.  not used here
        .do_out(adc_data),      // adc value from conversion
        .drdy_out(adc_data_ready),  //tells you valid data is ready to be latched
        .eoc_out(adc_ready),   //  specifies that the ADC is ready (conversion complete)
        .eos_out(eos_out),     //  specifies that conversion sequence is complete
        .alarm_out(alarm),      // OR's output of all internal alarms, not used here
        .vp_in(1'b0),           // dedicated analog input pair for differential, tied to 0 if not used
        .vn_in(1'b0)
    );
    //
    // wait for XADC to tell you something is ready to latch. note this means continuous latching
    //
    reg [15:0] r_adc_data;
    always @ (posedge adc_data_ready) begin
        if (reset) r_adc_data <= 16'h0;
        else r_adc_data <= adc_data;
        end
    //
    //  make a ~1Hz clock so we can run the LED display slower
    //
    reg [19:0] counter;
    reg [15:0] s_adc_data;
    always @ (posedge clock1) begin
        if (reset_1s) counter <= 0;
        else counter <= counter + 1;
        end
    wire clock_1hz = counter[19];
    always @ (posedge clock_1hz) s_adc_data <= r_adc_data;
    assign display_this = version_level ? VERSION : s_adc_data;
    //
    //  instantiate our fifo next
    //
    wire [17:0] din = r_adc_data;   // feed this directly into the FIFO data input
    wire [17:0] dout;
    wire wr_en = sw[0];             // set write enable to the bottom slide switch
    wire rd_en, fifo_full, fifo_empty, wr_rst_busy, rd_rst_busy;
    wire [9:0] fifo_data_count;
    FIFO_DATA ourfifo (
        .rst(reset_1s),
        .wr_clk(clock1),
        .rd_clk(clock25),
        .din(din),
        .wr_en(wr_en),
        .rd_en(rd_en),
        .dout(dout),
        .full(fifo_full),
        .empty(fifo_empty),
        .rd_data_count(fifo_data_count),
        .wr_rst_busy(wr_rst_busy),
        .rd_rst_busy(rd_rst_busy)
    );
    //
    //  now make a state machine to deal with transmitting 2 bytes at a
    //  time, including the FIFO value that is present.
    //
    reg [2:0] tx_state;     // 8 states so 3 bits will do
    localparam [2:0] TX_IDLE=0, TX_BYTE1=1, TX_DO1=2, TX_WAIT1=3, TX_BYTE2=4,
                    TX_DO2=5, TX_WAIT2=6, TX_DONE=7;
    reg doit;
    wire [16:0] transmit_word;
    wire begin_transfer;
    reg [7:0] tx_data;
    always @ (posedge clock25) begin
        if (reset) begin
            tx_state <= TX_IDLE;
            doit <= 0;
            tx_data <= 0;
        end
        else 
            case (tx_state)
                TX_IDLE: begin
                    //
                    //  wait for begin_transfer to start
                    //
                    if (begin_transfer) tx_state <= TX_BYTE1;
                    else tx_state <= TX_IDLE;
                    doit <= 0;
                    tx_data <= 0;
                end
                TX_BYTE1: begin
                    //
                    //  latch the first byte of the transmit_word
                    //
                    tx_data <= transmit_word[7:0];
                    tx_state <= TX_DO1;
                end
                TX_DO1: begin
                    //
                    //  turn on uart_tx
                    //
                    doit <= 1;
                    tx_state <= TX_WAIT1;
                end
                TX_WAIT1: begin
                    //
                    //  turn off doit and wait for uart_tx to finish
                    //
                    doit <= 0;
                    if (tx_done) tx_state <= TX_BYTE2;
                    else tx_state <= TX_WAIT1;
                end
                TX_BYTE2: begin
                    //
                    //  latch the 2nd byte to transfer
                    //
                    tx_data <= transmit_word[15:8];
                    tx_state <= TX_DO2;
                end
                TX_DO2: begin
                    //
                    //  turn on uart_tx
                    //
                    doit <= 1;
                    tx_state <= TX_WAIT2;
                end
                TX_WAIT2: begin
                    //
                    //  wait for the 2nd byte to finish begin sent
                    //
                    doit <= 0;
                    if (tx_done) tx_state <= TX_DONE;
                    else tx_state <= TX_WAIT2;
                end
                TX_DONE: begin
                    //
                    //  wait for begin_transfer to go away
                    //
                    if (do_transmit) tx_state <= TX_DONE;
                    else tx_state <= TX_IDLE;
                end
                default: begin
                    tx_state <= TX_IDLE;
                end
            endcase
    end
    assign do_transmit = doit;
    assign transmit_byte = tx_data;
    //
    //  we now have a XADC and FIFO that run on the 1MHz clock1 clock
    //  (the FIFO write, that is).  we can read the FIFO with the 25MHz
    //  clock that runs on the FSM next.
    //
    //  next comes the uart receive state machine. we used the bottom few
    //  bits of the incoming byte from the RPi:
    //
    //  7654 3210
    //
    //  0000 0001   read 1 word from the fifo and send back 16 bits of data
    //  0000 0011   transmit r_adc_data
    //  0000 0101   transmit slide switches sw
    //  0000 0111   transmit firmware version VERSION
    //  0000 1001   read 1 word from the fifo and send back data count
    //
    //  when we send the fifo data back, just keep the upper 12 bits from
    //  the ADC value, and add the following 4 bits at the bottom:
    //
    //  0,fifo_full,fifo_empty,fifo_last
    //
    //  where fifo_last is asserted if fifo_data_count == 1, which
    //  means this is the last word in the fifo.
    //
    reg [3:0] rx_state;
    localparam [3:0] RX_WAIT=0, RX_INCOMING=1, RX_LATCH=2, RX_SEND=3,
       RX_WAIT_DONE1=4, RX_PAUSE=5, RX_WAIT_DONE2=6, RX_READ_FIFO=7,
       RX_WAIT_FIFO_DONE=8, RX_FIFO_LATCH=9;
    reg start_tx;                                       // tells TX FSM to start sending 2 bytes
    reg [15:0] transmit_this;                           // 16 bits latched, goes to TX FSM
    reg fifo_rd;                                        // tells FIFO FSM to read 1 word
    reg [7:0] received_byte;                            // latch what comes out of uart_rx
    wire received_request = received_byte[0];                 // this has to be on for something to be sent back
    wire [2:0] latch_what = received_byte[3:1];         // what does the RPi want to be sent back?
    wire send_fifo_data = (latch_what == 3'b000);       // send fifo data
    wire send_fifo_count = (latch_what == 3'b100);
    localparam [2:0]    SEND_FIFO_DATA=3'b000, 
                        SEND_1_ADC=3'b001,
                        SEND_TEST=3'b010,
                        SEND_VERSION=3'b011,
                        SEND_FIFO_COUNT=3'b100;
    wire send_fifo = send_fifo_data | send_fifo_count;  // either of the above
    wire incoming = dv;                                 // signal to get going
    wire fifo_done;                                     // comes out of f_state below
    wire [15:0] fifo_data_latched;                      // ditto
    wire [15:0] fifo_count_latched;                     // ditto
    always @ (posedge clock25) begin
        if (reset) begin
            start_tx <= 0;
            rx_state <= RX_WAIT;
            transmit_this <= 0;
            fifo_rd <= 0;
            received_byte <= rx_data;
        end
        else 
            case (rx_state)
                RX_WAIT: begin
                    start_tx <= 0;
                    transmit_this <= 0;
                    fifo_rd <= 0;
                    received_byte <= rx_data;
                    if (incoming) rx_state <= RX_INCOMING;
                    else rx_state <= RX_WAIT;
                end
                RX_INCOMING: begin
                    if (received_request) begin
                        if (send_fifo) rx_state <= RX_READ_FIFO;
                        else rx_state <= RX_LATCH;
                    end
                    else rx_state <= RX_WAIT;
                end
                RX_LATCH: begin
                    rx_state <= RX_SEND;
                    case (latch_what) 
                        SEND_1_ADC: transmit_this <= r_adc_data;
                        SEND_TEST: transmit_this <= sw;
                        SEND_VERSION: transmit_this <= VERSION;
                        default: transmit_this <= 'heeee;
                    endcase
                end
                RX_READ_FIFO: begin
                    //
                    //  read the fifo for 1 clock tick
                    //
                    fifo_rd <= 1;
                    rx_state <= RX_WAIT_FIFO_DONE;
                end
                RX_WAIT_FIFO_DONE: begin
                    if (fifo_done) rx_state <= RX_FIFO_LATCH;
                    else rx_state <= RX_WAIT_FIFO_DONE;
                end
                RX_FIFO_LATCH: begin
                    transmit_this <= send_fifo_data ? fifo_data_latched : fifo_count_latched;
                    rx_state <= RX_SEND;
                end
                RX_SEND: begin
                    fifo_rd <= 0;
                    start_tx <= 1;
                    rx_state <= RX_WAIT_DONE1;                    
                end
                RX_WAIT_DONE1: begin
                    if (tx_done) rx_state <= RX_PAUSE;
                    else rx_state <= RX_WAIT_DONE1;
                end
                RX_PAUSE: begin
                    rx_state <= RX_WAIT_DONE2;
                end
                RX_WAIT_DONE2: begin
                    if (tx_done) rx_state <= RX_WAIT_DONE2;
                    else rx_state <= RX_WAIT;
                end
            endcase
    end
    assign transmit_word = transmit_this;
    assign begin_transfer = start_tx;
    //
    //  this FSM controlls the FIFO reading
    //
    localparam [1:0] FIFO_IDLE=0, FIFO_READ=1, FIFO_LATCH=2, FIFO_WAIT_DONE=3;
    reg [1:0] f_state;
    reg [15:0] f_data_out;
    reg [16:0] f_count_out;
    reg f_rd_en;
    reg f_done;
    wire fifo_last = (fifo_data_count == 10'h1);
    wire [3:0] bottom_bits = {1'b0,fifo_full,fifo_empty,fifo_last};
    always @ (posedge clock25)
        if (reset) begin
            f_state <= FIFO_IDLE;
            f_data_out <= 0;
            f_count_out <= 0;
            f_done <= 0;
            f_rd_en <= 0;
        end
        else case (f_state)
            FIFO_IDLE: begin
                f_data_out <= 0;
                f_count_out <= 0;
                f_done <= 0;
                f_rd_en <= 0;
                if (fifo_rd) f_state <= FIFO_READ;
                else f_state <= FIFO_IDLE;
            end
            FIFO_READ: begin
                f_rd_en <= 1;
                f_state <= FIFO_LATCH;
            end
            FIFO_LATCH: begin
                f_rd_en <= 0;
                f_data_out <= {dout[15:4], bottom_bits[3:0]};
                f_count_out <= {6'h0,fifo_data_count[9:0]};
                f_state <= FIFO_WAIT_DONE;
            end
            FIFO_WAIT_DONE: begin
                f_done <= 1;
                if (fifo_rd) f_state <= FIFO_WAIT_DONE;
                else f_state <= FIFO_IDLE;
            end
        endcase
    assign rd_en = f_rd_en;
    assign fifo_data_latched = f_data_out;
    assign fifo_count_latched = f_count_out;
    assign fifo_done = f_done;
    //
    assign led = {6'b0,rx_state[3:0],tx_state[2:0],fifo_empty,fifo_full,sw[0]};
    always @ (*) 
        case (sw[15:13])
            3'b000: begin
                JB = {rd_en,tx_done,tx,do_transmit,send_fifo_data,rx,clock25,clock1};
                JA = {f_rd_en,f_state[1:0],fifo_rd,rx_state[3:0]};
            end
            3'b001: begin
                JB = {adc_ready,adc_data_ready,isbusy,eos_out,adc_data[3:0]};
                JA = {rd_en,dout[6:0]};
            end
            3'b010: begin
                JB = received_byte[7:0];
                JA = {f_rd_en,f_state[1:0],fifo_rd,rx_state[3:0]};
            end
            3'b011: begin
                JB = {dv,rx_data[6:0]};
                JA = {f_rd_en,f_state[1:0],fifo_rd,rx_state[3:0]};
            end
            3'b100: begin
                JB = rx_data[7:0];
                JA = received_byte[7:0];
            end
            default: begin
                JB = {rd_en,tx_done,tx,do_transmit,send_fifo_data,rx,clock25,clock1};
                JA = {f_rd_en,f_state[1:0],fifo_rd,rx_state[3:0]};
            end
        endcase
/*
    assign JB =  sw[14] ?
        {rd_en,tx_done,tx,do_transmit,send_fifo_data,rx,clock25,clock1} :
        {f_rd_en,f_state[1:0],fifo_rd,rx_state[3:0]};
    assign JA = sw[15] ? 
        {adc_ready,adc_data_ready,isbusy,eos_out,adc_data[3:0]} :
        {rd_en,dout[6:0]};
*/
endmodule
